@@ -33,7 +33,7 @@ import java.util.stream.Stream;
 import io.airlift.airline.Command;
 import org.apache.activemq.artemis.util.JVMArgumentParser;
 
-@Command(name = "upgrade", description = "Update an artemis instance to the current artemis.home, keeping all the data and broker.xml. Warning: backup your instance before using this command and compare the files.")
+@Command(name = "upgrade", description = "Update a broker instance to the current artemis.home, keeping all the data and broker.xml. Warning: backup your instance before using this command and compare the files.")
 public class Upgrade extends InstallAbstract {
 
    // these are the JVM argumnents we must keep between upgrades
@@ -182,7 +182,12 @@ public class Upgrade extends InstallAbstract {
       final File bootstrapXmlBkp = new File(etcBkp, Create.ETC_BOOTSTRAP_XML);
 
       Files.copy(bootstrapXml.toPath(), bootstrapXmlTmp.toPath());
-      replaceLines(context, bootstrapXmlTmp, bootstrapXml, bootstrapXmlBkp, "<web path", "   <web path=\"web\" rootRedirectLocation=\"console\">");
+      replaceLines(context, bootstrapXmlTmp, bootstrapXml, bootstrapXmlBkp,
+         "^(.*)<web path.*$", "$1<web path=\"web\" rootRedirectLocation=\"console\">",
+         "^(.*)<binding uri=\"http://localhost:8161\"(.*)$", "$1<binding name=\"artemis\" uri=\"http://localhost:8161\"$2",
+         "^(.*)<app(.*branding.*)$", "$1<app name=\"branding\"$2",
+         "^(.*)<app(.*plugin.*)$", "$1<app name=\"plugin\"$2",
+         "^(.*)<app url=\"([^\"]+)\"(.*)$", "$1<app name=\"$2\" url=\"$2\"$3");
 
       upgradeLogging(context, etcFolder, etcBkp);
 
@@ -258,8 +263,8 @@ public class Upgrade extends InstallAbstract {
                 null,
                 newLine -> {
                    for (int i = 0; i < replacePairs.length; i += 2) {
-                      if (newLine.trim().startsWith(replacePairs[i])) {
-                         return replacePairs[i + 1];
+                      if (newLine.matches(replacePairs[i])) {
+                         return newLine.replaceAll(replacePairs[i], replacePairs[i + 1]);
                       }
                    }
                    return newLine;
@@ -340,7 +345,7 @@ public class Upgrade extends InstallAbstract {
          if (!newLogging.exists()) {
             context.out.println("Creating " + newLogging);
             try (InputStream inputStream = openStream("etc/" + Create.ETC_LOG4J2_PROPERTIES);
-                 OutputStream outputStream = new FileOutputStream(newLogging);) {
+                 OutputStream outputStream = new FileOutputStream(newLogging)) {
                copy(inputStream, outputStream);
             }
          }
