@@ -1,12 +1,12 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
- * contributor license agreements. See the NOTICE file distributed with
+ * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
  * The ASF licenses this file to You under the Apache License, Version 2.0
  * (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
+ * the License.  You may obtain a copy of the License at
  *
- * http://www.apache.org/licenses/LICENSE-2.0
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
@@ -14,7 +14,6 @@
  * See the License for the specific language governing permissions and
  * limitations under the License.
  */
-
 package org.apache.activemq.artemis.core.protocol.mqtt;
 
 import java.util.Arrays;
@@ -43,10 +42,14 @@ import org.apache.activemq.artemis.core.io.IOCallback;
 import org.apache.activemq.artemis.core.protocol.mqtt.exceptions.DisconnectException;
 import org.apache.activemq.artemis.core.server.Queue;
 import org.apache.activemq.artemis.core.server.ServerConsumer;
+import org.apache.activemq.artemis.core.server.ServerProducer;
+import org.apache.activemq.artemis.core.server.impl.AddressInfo;
 import org.apache.activemq.artemis.core.server.impl.ServerSessionImpl;
 import org.apache.activemq.artemis.core.transaction.Transaction;
-import org.apache.activemq.artemis.reader.MessageUtil;
-import org.jboss.logging.Logger;
+import org.apache.activemq.artemis.utils.UUIDGenerator;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
+import java.lang.invoke.MethodHandles;
 
 import static io.netty.handler.codec.mqtt.MqttProperties.MqttPropertyType.CONTENT_TYPE;
 import static io.netty.handler.codec.mqtt.MqttProperties.MqttPropertyType.CORRELATION_DATA;
@@ -68,16 +71,18 @@ import static org.apache.activemq.artemis.core.protocol.mqtt.MQTTUtil.MQTT_USER_
  */
 public class MQTTPublishManager {
 
-   private static final Logger logger = Logger.getLogger(MQTTPublishManager.class);
+   private static final Logger logger = LoggerFactory.getLogger(MethodHandles.lookup().lookupClass());
 
    private SimpleString managementAddress;
 
+   private final String senderName = UUIDGenerator.getInstance().generateUUID().toString();
+
+   private boolean createProducer = true;
+
    private ServerConsumer managementConsumer;
 
    private MQTTSession session;
 
-   private MQTTLogger log = MQTTLogger.LOGGER;
-
    private final Object lock = new Object();
 
    private MQTTSessionState state;
@@ -175,6 +180,10 @@ public class MQTTPublishManager {
     */
    void sendToQueue(MqttPublishMessage message, boolean internal) throws Exception {
       synchronized (lock) {
+         if (createProducer) {
+            session.getServerSession().addProducer(senderName, MQTTProtocolManagerFactory.MQTT_PROTOCOL_NAME, ServerProducer.ANONYMOUS);
+            createProducer = false;
+         }
          String topic = message.variableHeader().topicName();
          if (session.getVersion() == MQTTVersion.MQTT_5) {
             Integer alias = MQTTUtil.getProperty(Integer.class, message.variableHeader().properties(), TOPIC_ALIAS);
@@ -190,6 +199,10 @@ public class MQTTPublishManager {
                   topic = session.getState().getClientTopicAlias(alias);
                   if (topic == null) {
                      topic = message.variableHeader().topicName();
+                     if (topic == null || topic.length() == 0) {
+                        // using a topic alias with no matching topic in the state; potentially [MQTT-3.3.2-7]
+                        throw new DisconnectException(MQTTReasonCodes.TOPIC_ALIAS_INVALID);
+                     }
                      session.getState().addClientTopicAlias(alias, topic);
                   }
                }
@@ -197,13 +210,13 @@ public class MQTTPublishManager {
                topic = message.variableHeader().topicName();
             }
          }
-
-         Message serverMessage = MQTTUtil.createServerMessageFromByteBuf(session, topic, message);
+         String coreAddress = MQTTUtil.convertMqttTopicFilterToCoreAddress(topic, session.getWildcardConfiguration());
+         SimpleString address = SimpleString.toSimpleString(coreAddress, session.getCoreMessageObjectPools().getAddressStringSimpleStringPool());
+         Message serverMessage = MQTTUtil.createServerMessageFromByteBuf(session, address, message);
          int qos = message.fixedHeader().qosLevel().value();
          if (qos > 0) {
             serverMessage.setDurable(MQTTUtil.DURABLE_MESSAGES);
          }
-
          int messageId = message.variableHeader().packetId();
          if (qos < 2 || !state.getPubRec().contains(messageId)) {
             if (qos == 2 && !internal)
@@ -211,11 +224,15 @@ public class MQTTPublishManager {
 
             Transaction tx = session.getServerSession().newTransaction();
             try {
-               if (internal) {
-                  session.getServer().getPostOffice().route(serverMessage, tx, true);
-               } else {
-                  session.getServerSession().send(tx, serverMessage, true, false);
+               AddressInfo addressInfo = session.getServer().getAddressInfo(address);
+               if (addressInfo == null && session.getServer().getAddressSettingsRepository().getMatch(coreAddress).isAutoCreateAddresses()) {
+                  session.getServerSession().createAddress(address, RoutingType.MULTICAST, true);
+                  serverMessage.setRoutingType(RoutingType.MULTICAST);
                }
+               if (addressInfo != null) {
+                  serverMessage.setRoutingType(addressInfo.getRoutingType());
+               }
+               session.getServerSession().send(tx, serverMessage, true, senderName, false);
 
                if (message.fixedHeader().isRetain()) {
                   ByteBuf payload = message.payload();
@@ -225,6 +242,9 @@ public class MQTTPublishManager {
                tx.commit();
             } catch (ActiveMQSecurityException e) {
                tx.rollback();
+               if (internal) {
+                  throw e;
+               }
                if (session.getVersion() == MQTTVersion.MQTT_5) {
                   sendMessageAck(internal, qos, messageId, MQTTReasonCodes.NOT_AUTHORIZED);
                   return;
@@ -241,9 +261,7 @@ public class MQTTPublishManager {
                   if (closeMqttConnectionOnPublishAuthorizationFailure) {
                      throw e;
                   } else {
-                     if (logger.isDebugEnabled()) {
-                        logger.debug("MQTT 3.1.1 client not authorized to publish message.");
-                     }
+                     logger.debug("MQTT 3.1.1 client not authorized to publish message.");
                   }
                } else {
                   /*
@@ -256,17 +274,16 @@ public class MQTTPublishManager {
                    *
                    * Log the failure since we have to just swallow it.
                    */
-                  if (logger.isDebugEnabled()) {
-                     logger.debug("MQTT 3.1 client not authorized to publish message.");
-                  }
+                  logger.debug("MQTT 3.1 client not authorized to publish message.");
                }
             } catch (Throwable t) {
                MQTTLogger.LOGGER.failedToPublishMqttMessage(t.getMessage(), t);
                tx.rollback();
                throw t;
             }
-            createMessageAck(messageId, qos, internal);
          }
+
+         createMessageAck(messageId, qos, internal);
       }
    }
 
@@ -301,7 +318,7 @@ public class MQTTPublishManager {
          if (ref != null) {
             Message m = MQTTUtil.createPubRelMessage(session, getManagementAddress(), messageId);
             //send the management message via the internal server session to bypass security.
-            session.getInternalServerSession().send(m, true);
+            session.getInternalServerSession().send(m, true, senderName);
             session.getServerSession().individualAcknowledge(ref.getB(), ref.getA());
             releaseFlowControl(ref.getB());
          } else {
@@ -341,7 +358,7 @@ public class MQTTPublishManager {
 
          @Override
          public void onError(int errorCode, String errorMessage) {
-            log.error("Pub Sync Failed");
+            logger.error("Pub Sync Failed");
          }
       });
    }
@@ -361,7 +378,7 @@ public class MQTTPublishManager {
             releaseFlowControl(ref.getB());
          }
       } catch (ActiveMQIllegalStateException e) {
-         log.warn("MQTT Client(" + session.getState().getClientId() + ") attempted to Ack already Ack'd message");
+         logger.warn("MQTT Client({}) attempted to Ack already Ack'd message", session.getState().getClientId());
       }
    }
 
@@ -395,9 +412,6 @@ public class MQTTPublishManager {
             isRetain = false;
          }
 
-         // [MQTT-3.8.3-3] remove property used for no-local implementation
-         message.removeProperty(MessageUtil.CONNECTION_ID_PROPERTY_NAME);
-
          if (session.getState().getClientTopicAliasMaximum() != null) {
             Integer alias = session.getState().getServerTopicAlias(address);
             if (alias == null) {
@@ -425,7 +439,7 @@ public class MQTTPublishManager {
              * [MQTT-3.1.2-25] Where a Packet is too large to send, the Server MUST discard it without sending it and then
              * behave as if it had completed sending that Application Message
              */
-            logger.debugf("Not sending message %s to client as its size (%d) exceeds the max (%d)", message, size, maxSize);
+            logger.debug("Not sending message {} to client as its size ({}) exceeds the max ({})", message, size, maxSize);
             session.getServerSession().individualAcknowledge(consumerId, message.getMessageID());
             return false;
          }
