@@ -16,18 +16,6 @@
  */
 package org.apache.activemq.artemis.core.protocol.openwire;
 
-import java.io.IOException;
-import java.util.ArrayList;
-import java.util.List;
-import java.util.ListIterator;
-import java.util.Map;
-import java.util.Set;
-import java.util.concurrent.ConcurrentHashMap;
-import java.util.concurrent.Executor;
-import java.util.concurrent.TimeUnit;
-import java.util.concurrent.atomic.AtomicBoolean;
-import java.util.concurrent.atomic.AtomicLongFieldUpdater;
-
 import javax.jms.IllegalStateException;
 import javax.jms.InvalidClientIDException;
 import javax.jms.InvalidDestinationException;
@@ -35,16 +23,28 @@ import javax.jms.JMSSecurityException;
 import javax.transaction.xa.XAException;
 import javax.transaction.xa.XAResource;
 import javax.transaction.xa.Xid;
+import java.io.IOException;
+import java.lang.invoke.MethodHandles;
+import java.util.ArrayList;
+import java.util.List;
+import java.util.ListIterator;
+import java.util.Map;
+import java.util.Set;
+import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.Executor;
+import java.util.concurrent.ScheduledFuture;
+import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicLongFieldUpdater;
 
 import org.apache.activemq.advisory.AdvisorySupport;
-import org.apache.activemq.artemis.api.core.ActiveMQAddressExistsException;
 import org.apache.activemq.artemis.api.core.ActiveMQBuffer;
 import org.apache.activemq.artemis.api.core.ActiveMQException;
 import org.apache.activemq.artemis.api.core.ActiveMQExceptionType;
 import org.apache.activemq.artemis.api.core.ActiveMQNonExistentQueueException;
-import org.apache.activemq.artemis.api.core.ActiveMQQueueExistsException;
 import org.apache.activemq.artemis.api.core.ActiveMQRemoteDisconnectException;
 import org.apache.activemq.artemis.api.core.ActiveMQSecurityException;
+import org.apache.activemq.artemis.api.core.AutoCreateResult;
 import org.apache.activemq.artemis.api.core.QueueConfiguration;
 import org.apache.activemq.artemis.api.core.RoutingType;
 import org.apache.activemq.artemis.api.core.SimpleString;
@@ -63,19 +63,14 @@ import org.apache.activemq.artemis.core.protocol.openwire.util.OpenWireUtil;
 import org.apache.activemq.artemis.core.remoting.FailureListener;
 import org.apache.activemq.artemis.core.remoting.impl.netty.TransportConstants;
 import org.apache.activemq.artemis.core.security.SecurityAuth;
-import org.apache.activemq.artemis.core.server.ActiveMQMessageBundle;
 import org.apache.activemq.artemis.core.server.ActiveMQServer;
 import org.apache.activemq.artemis.core.server.ActiveMQServerLogger;
-import org.apache.activemq.artemis.core.server.AddressQueryResult;
-import org.apache.activemq.artemis.core.server.QueueQueryResult;
 import org.apache.activemq.artemis.core.server.MessageReference;
 import org.apache.activemq.artemis.core.server.ServerConsumer;
 import org.apache.activemq.artemis.core.server.ServerSession;
 import org.apache.activemq.artemis.core.server.SlowConsumerDetectionListener;
 import org.apache.activemq.artemis.core.server.TempQueueObserver;
-import org.apache.activemq.artemis.core.server.impl.AddressInfo;
 import org.apache.activemq.artemis.core.server.impl.RefsOperation;
-import org.apache.activemq.artemis.core.settings.impl.AddressSettings;
 import org.apache.activemq.artemis.core.transaction.ResourceManager;
 import org.apache.activemq.artemis.core.transaction.Transaction;
 import org.apache.activemq.artemis.core.transaction.TransactionOperationAbstract;
@@ -135,7 +130,6 @@ import org.apache.activemq.transport.TransmitCallback;
 import org.apache.activemq.util.ByteSequence;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
-import java.lang.invoke.MethodHandles;
 
 /**
  * Represents an activemq connection.
@@ -143,6 +137,8 @@ import java.lang.invoke.MethodHandles;
 public class OpenWireConnection extends AbstractRemotingConnection implements SecurityAuth, TempQueueObserver {
 
 
+   private final Object lockSend = new Object();
+
    // to be used on the packet size estimate processing for the ThresholdActor
    private static final int MINIMAL_SIZE_ESTIAMTE = 1024;
 
@@ -154,6 +150,8 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
 
    private boolean destroyed = false;
 
+   private volatile ScheduledFuture ttlCheck;
+
    //separated in/out wireFormats allow deliveries (eg async and consumers) to not slow down bufferReceived
    private final OpenWireFormat inWireFormat;
 
@@ -548,12 +546,16 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
          final int bufferSize = bytes.length;
          final int maxChunkSize = protocolManager.getOpenwireMaxPacketChunkSize();
 
-         if (maxChunkSize > 0 && bufferSize > maxChunkSize) {
-            chunkSend(bytes, bufferSize, maxChunkSize);
-         } else {
-            final ActiveMQBuffer buffer = transportConnection.createTransportBuffer(bufferSize);
-            buffer.writeBytes(bytes.data, bytes.offset, bufferSize);
-            transportConnection.write(buffer, false, false);
+         // We can't let any other packet to sneak in while chunkSend is happening.
+         // otherwise we may get wrong packts delivered
+         synchronized (lockSend) {
+            if (maxChunkSize > 0 && bufferSize > maxChunkSize) {
+               chunkSend(bytes, bufferSize, maxChunkSize);
+            } else {
+               final ActiveMQBuffer buffer = transportConnection.createTransportBuffer(bufferSize);
+               buffer.writeBytes(bytes.data, bytes.offset, bufferSize);
+               transportConnection.write(buffer, false, false);
+            }
          }
          bufferSent();
       } catch (IOException e) {
@@ -575,7 +577,7 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
          }
          final ActiveMQBuffer chunk = transportConnection.createTransportBuffer(chunkSize);
          chunk.writeBytes(bytes.data, bytes.offset, chunkSize);
-         transportConnection.write(chunk, true, false);
+         transportConnection.write(chunk, false, false);
          bytes.setOffset(bytes.getOffset() + chunkSize);
       }
    }
@@ -682,10 +684,18 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
 
    private void shutdown(boolean fail) {
 
-      if (fail) {
-         transportConnection.forceClose();
-      } else {
-         transportConnection.close();
+      try {
+         if (fail) {
+            transportConnection.forceClose();
+         } else {
+            transportConnection.close();
+         }
+      } finally {
+         ScheduledFuture ttlCheckToCancel = this.ttlCheck;
+         this.ttlCheck = null;
+         if (ttlCheckToCancel != null) {
+            ttlCheckToCancel.cancel(true);
+         }
       }
    }
 
@@ -837,7 +847,7 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
    }
 
    private void createInternalSession(ConnectionInfo info) throws Exception {
-      internalSession = server.createSession(UUIDGenerator.getInstance().generateStringUUID(), context.getUserName(), info.getPassword(), -1, this, true, false, false, false, null, null, true, operationContext, protocolManager.getPrefixes(), protocolManager.getSecurityDomain(), validatedUser);
+      internalSession = server.createSession(UUIDGenerator.getInstance().generateStringUUID(), context.getUserName(), info.getPassword(), -1, this, true, false, false, false, null, null, true, operationContext, protocolManager.getPrefixes(), protocolManager.getSecurityDomain(), validatedUser, false);
    }
 
    //raise the refCount of context
@@ -889,29 +899,18 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
       }
 
       SimpleString qName = SimpleString.toSimpleString(dest.getPhysicalName());
-      if (server.locateQueue(qName) == null) {
-         AddressSettings addressSettings = server.getAddressSettingsRepository().getMatch(dest.getPhysicalName());
-         if (dest.isQueue() && (addressSettings.isAutoCreateQueues() || dest.isTemporary())) {
-            try {
-               internalSession.createQueue(new QueueConfiguration(qName).setRoutingType(RoutingType.ANYCAST).setDurable(!dest.isTemporary()).setTemporary(dest.isTemporary()).setAutoCreated(!dest.isTemporary()));
-               created = true;
-            } catch (ActiveMQQueueExistsException exists) {
-               // The queue may have been created by another thread in the mean time.  Catch and do nothing.
-            }
-         } else if (dest.isTopic() && (addressSettings.isAutoCreateAddresses() || dest.isTemporary())) {
-            try {
-               AddressInfo addressInfo = new AddressInfo(qName, RoutingType.MULTICAST);
-               if (AdvisorySupport.isAdvisoryTopic(dest) && protocolManager.isSuppressInternalManagementObjects()) {
-                  addressInfo.setInternal(true);
-               }
-               if (internalSession.getAddress(addressInfo.getName()) == null) {
-                  internalSession.createAddress(addressInfo, !dest.isTemporary());
-                  created = true;
-               }
-            } catch (ActiveMQAddressExistsException exists) {
-               // The address may have been created by another thread in the mean time.  Catch and do nothing.
-            }
+
+      AutoCreateResult autoCreateResult = internalSession.checkAutoCreate(new QueueConfiguration(qName)
+                                                                             .setRoutingType(dest.isQueue() ? RoutingType.ANYCAST : RoutingType.MULTICAST)
+                                                                             .setDurable(!dest.isTemporary())
+                                                                             .setTemporary(dest.isTemporary()));
+      if (autoCreateResult == AutoCreateResult.CREATED) {
+         created = true;
+         if (AdvisorySupport.isAdvisoryTopic(dest) && protocolManager.isSuppressInternalManagementObjects()) {
+            internalSession.getAddress(qName).setInternal(true);
          }
+      } else if (autoCreateResult == AutoCreateResult.NOT_FOUND) {
+         throw new InvalidDestinationException(dest.getDestinationTypeAsString() + " " + dest.getPhysicalName() + " does not exist.");
       }
 
       if (dest.isTemporary()) {
@@ -1008,7 +1007,7 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
       this.maxInactivityDuration = inactivityDuration;
 
       if (this.useKeepAlive) {
-         protocolManager.getScheduledPool().schedule(new Runnable() {
+         ttlCheck = protocolManager.getScheduledPool().schedule(new Runnable() {
             @Override
             public void run() {
                if (inactivityDuration >= 0) {
@@ -1155,23 +1154,6 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
       }
    }
 
-   /**
-    * Checks to see if this destination exists.  If it does not throw an invalid destination exception.
-    *
-    * @param destination
-    */
-   private void validateDestination(ActiveMQDestination destination) throws Exception {
-      if (destination.isQueue()) {
-         SimpleString physicalName = new SimpleString(destination.getPhysicalName());
-         QueueQueryResult queue = server.queueQuery(physicalName);
-         AddressQueryResult address = server.addressQuery(physicalName);
-
-         if (!address.isExists() && !queue.isAutoCreateQueues()) {
-            throw ActiveMQMessageBundle.BUNDLE.noSuchQueue(physicalName);
-         }
-      }
-   }
-
    private void propagateLastSequenceId(SessionState sessionState, long lastDeliveredSequenceId) {
       for (ConsumerState consumerState : sessionState.getConsumerStates()) {
          consumerState.getInfo().setLastDeliveredSequenceId(lastDeliveredSequenceId);
@@ -1253,14 +1235,14 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
             ActiveMQDestination destination = info.getDestination();
 
             if (destination != null && !AdvisorySupport.isAdvisoryTopic(destination)) {
-               if (destination.isQueue()) {
-                  OpenWireConnection.this.validateDestination(destination);
-               }
-               DestinationInfo destInfo = new DestinationInfo(getContext().getConnectionId(), DestinationInfo.ADD_OPERATION_TYPE, destination);
-               OpenWireConnection.this.addDestination(destInfo);
+               OpenWireConnection.this.addDestination(new DestinationInfo(getContext().getConnectionId(), DestinationInfo.ADD_OPERATION_TYPE, destination));
             }
 
             ss.addProducer(info);
+            getSession(info.getProducerId().getParentId()).getCoreSession().addProducer(
+                  info.getProducerId().toString(),
+                  OpenWireProtocolManagerFactory.OPENWIRE_PROTOCOL_NAME,
+                  info.getDestination() != null ? info.getDestination().getPhysicalName() : null);
 
          }
          return null;
@@ -1290,6 +1272,7 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
          }
          synchronized (producerExchanges) {
             producerExchanges.remove(id);
+            getSession(id.getParentId()).getCoreSession().removeProducer(id.toString());
          }
          return null;
       }
@@ -1803,7 +1786,9 @@ public class OpenWireConnection extends AbstractRemotingConnection implements Se
    }
 
    private void clearupOperationContext() {
-      server.getStorageManager().clearContext();
+      if (server != null && server.getStorageManager() != null) {
+         server.getStorageManager().clearContext();
+      }
    }
 
    private Transaction lookupTX(TransactionId txID, AMQSession session) throws Exception {
