@@ -27,6 +27,7 @@ import java.io.DataOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
+import java.lang.invoke.MethodHandles;
 import java.util.Arrays;
 import java.util.Map;
 import java.util.Map.Entry;
@@ -75,12 +76,16 @@ import org.apache.activemq.util.ByteSequenceData;
 import org.apache.activemq.util.MarshallingSupport;
 import org.apache.activemq.wireformat.WireFormat;
 import org.fusesource.hawtbuf.UTF8Buffer;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
 
 import static org.apache.activemq.artemis.api.core.Message.HDR_INGRESS_TIMESTAMP;
 import static org.apache.activemq.command.ActiveMQDestination.QUEUE_TYPE;
 
 public final class OpenWireMessageConverter {
 
+   private static final Logger logger = LoggerFactory.getLogger(MethodHandles.lookup().lookupClass());
+
    public OpenWireMessageConverter() {
 
    }
@@ -106,9 +111,13 @@ public final class OpenWireMessageConverter {
       final ActiveMQBuffer body = coreMessage.getBodyBuffer();
 
       final ByteSequence contents = messageSend.getContent();
-      if (contents == null && coreType == org.apache.activemq.artemis.api.core.Message.TEXT_TYPE) {
-         body.writeNullableString(null);
-      } else if (contents != null) {
+      if (contents == null) {
+         if (coreType == org.apache.activemq.artemis.api.core.Message.TEXT_TYPE) {
+            body.writeNullableString(null);
+         } else if (coreType == org.apache.activemq.artemis.api.core.Message.MAP_TYPE) {
+            body.writeByte(DataConstants.NULL);
+         }
+      } else {
          final boolean messageCompressed = messageSend.isCompressed();
          if (messageCompressed) {
             coreMessage.putBooleanProperty(OpenWireConstants.AMQ_MSG_COMPRESSED, true);
@@ -258,7 +267,7 @@ public final class OpenWireMessageConverter {
          int n = ois.read(buf);
          while (n != -1) {
             decompressed.write(buf, 0, n);
-            n = ois.read();
+            n = ois.read(buf);
          }
          //read done
          return decompressed.toByteSequence();
@@ -860,7 +869,8 @@ public final class OpenWireMessageConverter {
       byte[] bytes = new byte[n];
       buffer.readBytes(bytes);
       if (isCompressed) {
-         try (ByteArrayOutputStream bytesOut = new ByteArrayOutputStream(); DeflaterOutputStream out = new DeflaterOutputStream(bytesOut, true)) {
+         try (ByteArrayOutputStream bytesOut = new ByteArrayOutputStream();
+              DeflaterOutputStream out = new DeflaterOutputStream(bytesOut, true)) {
             out.write(bytes);
             out.flush();
             bytes = bytesOut.toByteArray();
@@ -927,6 +937,10 @@ public final class OpenWireMessageConverter {
                                                  final Set<SimpleString> props) throws IOException {
       for (SimpleString s : props) {
          final String keyStr = s.toString();
+         if (keyStr.length() == 0) {
+            logger.debug("ignoring property with empty key name");
+            continue;
+         }
          if (!coreMessage.containsProperty(ManagementHelper.HDR_NOTIFICATION_TYPE) && (keyStr.startsWith("_AMQ") || keyStr.startsWith("__HDR_"))) {
             continue;
          }
